---
title: Introduction
date: 2022-12-18
hide:
  - footer
---

# Pseudo-3D cubes from densely spaced subbottom profiles via _Projection Onto Convex Sets_ (POCS) interpolation

This repository contains the source code for the generation of pseudo-3D cubes from densely spaced 2D subbottom profiles by utilizing the _Projection Onto Convex Sets_ ([POCS](./3D/3D_cube_interpolation/#pocs-theory)) method.

The source code accompanies a journal article published in ==[Geophysical Journal International]()==:

    Warnke et al. (2023) Geophysical Journal International

!!! info "Usage information"

    When using this workflow please refer to the [citation information](citation.md) and, in case of feedback or questions, the [contact information](contact.md).

## Installation

This worfklow can be installed in two ways:

### A) From GitHub repository

The workflow packages can be installed directly from GitHub:

```bash
>>> pip install git+https://github.com/fwrnke/pseudo-3D-interpolation.git
```

This installs all required dependencies and the workflow packages in your current (virtual) environment.

### B) Local installation

Alternatively, the workflow can be installed locally after downloading and unzipping the source code using the [ZIP archive](https://github.com/fwrnke/pseudo-3D-interpolation/archive/refs/heads/main.zip) from GitHub.

Navigate into the unzipped directory:

```bash
>>> cd ./pseudo-3D-interpolation  # root directory of unzipped package
```

Use the provided `conda` environment file to install dependencies (**optional**):

```bash
>>> conda install -f environment.yml  # install env
>>> conda activate pseudo_3d          # activate new env (default name: "pseudo_3d")
```

Install the workflow package in your active environment:

```bash
>>> pip install [-e] .  # -e: setuptools "develop mode"
```

The required dependencies will be installed automatically if they are not already available (i.e. previously installed using `conda`).

!!! warning

    The **Curvelet transform** is only available on Unix systems via the [`curvelops`](https://github.com/PyLops/curvelops) package!  
    Please refer to the [README](https://github.com/PyLops/curvelops#readme) of the project and [additional installation instructions](https://github.com/DIG-Kaust/conda_envs/blob/main/install_curvelops.sh) if you plan to use this transform.

## Workflow structure

This workflow was created to enable the interpolation of high-resolution subbottom profiler data (2-6 kHz) into pseudo-3D cubes. Potential applications include:

- **Pseudo-3D interpolation** of:
  
     - subbottom profiler data (SBP)
     - single channel reflection seismic (SCS)
     - multichannel reflection seismic (MCS)

- **Upsampling** of existing post-stack 3D seismic volumes (e.g., in crossline direction)

The workflow can be subdivided into two separate stages (**Figure 1**):

1. **2D (pre-)processing** of individual profiles
2. **3D interpolation** of binned, sparse cube

### 2D profile processing

The most essential pre-processing steps for a subsequent 3D interpolation aim to correct the vertical offset between the input profiles. Without a sufficient correction, such offsets, especially at profile intersections, result in low quality interpolation outputs. This issue is addressed by three processing steps in the fist stage (**Figure 1**, _upper panel_):

- [Static correction](2D/static_correction.md) (per profile)
- [Tidal heave compensation](2D/tide_compensation.md) (per profile)
- [Mistie correction](2D/mistie_correction.md) (using all profiles) 

!!! info "2D stage: Input data types"

    For most of the scripts of the 2D processing stage, three different input types are allowed:
    
    1. single **SEG-Y file** (e.g., `filename.sgy`)
    2. **datalist** of files to process (e.g., `datalist.txt`)
    3. **directory** with input files (e.g., `/input_dir`) 
    
    The `datalist` is a plain text file that contains the path and name of files to process relativ to the `datalist.txt` file:
    
        ./folder/file1.sgy
        ./folder/file2.sgy
    
    for an examplary folder structure like this:
    
        |-- folder/
        |    |-- file1.sgy
        |    |-- file2.sgy
        |-- datalist.txt

### 3D interpolation

In the second stage, the following processing steps are conducted (**Figure 1**, _lower panel_):

- [Binning and geometry setup](3D/cube_binning_geometry.md)
- [Preprocessing](3D/preprocessing_3D.md) of sparse 3D cube
     - [Gain application](3D/preprocessing_3D/#gain-application)
     - [Amplitude balancing](3D/preprocessing_3D/#trace-balancing)
     - [Frequency filter](3D/preprocessing_3D/#trace-balancing) (lowpass, highpass, bandpass)
     - [Trace resampling](3D/preprocessing_3D/#trace-resampling)
     - [Trace envelope](3D/preprocessing_3D/#trace-envelope)
- [Forward FFT](3D/apply_FFT_3D.md) (_time_ to _frequency_ domain)
- [Interpolation via Projection Onto Convex Sets (POCS) algorithm](3D/3D_cube_interpolation.md)
- [Inverse FFT](3D/apply_inverse_FFT_3D.md) (_frequency_ to _time_ domain)

!!! info "3D stage: Input data"

    Most scripts of the second, 3D processing and interpolation stage require a single `netCDF` file and, most often, one or more `YAML` configuration files.  
    An exception is the **binning** script, which needs multiple 2D **netCDF** files.

<figure markdown>
![Mistie (in time domain) for individual TOPAS profiles](./figures/Index_flowchart_full-workflow-schematics.png){ width="700" }
    <figcaption>Figure 1: Schematic diagram of pseudo-3D interpolation workflow showing two stages: 2D processing of each profile (upper panel) and 3D interpolation of sparse 3D cube (lower panel).</figcaption>
</figure>

!!! warning "Workflow applications"

    The workflow has only been tested on subbottom profiler data. However, additional _proof-of-concept_ tests will be conducted in the first half of 2023.

## Result examples

### Vertical offset corrections

**Figure 2** shows the effect of the different vertical offset correction methods applied to an example dataset of >200 profiles. The [static correction](2D/static_correction.md) mainly improves the reflection coherence of individual lines (**Figure 3**) with several acquisition artifacts remaining in the data (black arrows, **Figure 2**a, b). As the survey was conducted over multiple days, a [tide compensation](2D/tide_compensation.md) basically _detrends_ the individual profiles and improves the data quality (**Figure 2**c, d). Remaining [misties](2D/mistie_correction.md) are successfully corrected using a least-squares minimization approach (**Figure 2**e, f).

<figure markdown>
![Mistie (in time domain) for individual TOPAS profiles](./figures/Index_vertical_offsets.png){ width="700" }
    <figcaption>Figure 2: Vertical offsets (upper row) and effect on detected seafloor reflection (lower row) after static correction (a, b), tide compensation (c, d), and mistie correction (e, f), respectively.</figcaption>
</figure>

### 2D TOPAS processing

The following figure illustrate the capabilities of the first workflow stage _before_ (**Figure 3**, _upper row_) and _after_ applying all processing steps (**Figure 3**, _lower row_). Both the _full waveform_ and _envelope_ of the examplary profile section are displayed for comparison. Besides compensating the vertical offsets (_larger inset figures_), a [despiking](2D/2D_despike.md) algorithm was applied to remove abundant and random noise burst of 20 ms length (_smaller inset figures_).

<figure markdown>
![Mistie (in time domain) for individual TOPAS profiles](./figures/Index_TOPAS_processing.png){ width="500" }
    <figcaption>Figure 3: Examplary TOPAS profile displayed as unprocessed full-waveform (a) and envelope (b) as well as processed sections (c) and (d), respectively.</figcaption>
</figure>

### 3D interpolation results

The following figure illustrates (a) the sparse cube _before_ and (b) the full pseudo-3D cube _after_ [interpolation](3D/3D_cube_interpolation.md) via POCS algorithm.

<figure markdown>
![Mistie (in time domain) for individual TOPAS profiles](./figures/Index_sparse_interp_3D_edit.png){ width="800" }
    <figcaption>Figure 4: Sparse (a) and interpolated (b) pseudo-3D TOPAS cube overlain by multibeam bathymetry data.</figcaption>
</figure>

## Funding

This workflow was developed as part of my PhD research at the [School of Environment, University of Auckland, New Zealand](https://www.auckland.ac.nz/en/science/about-the-faculty/school-of-environment.html).

The funding was provided by the Royal Society of New Zealand Marsden Fund grant _Geologic champagne: What controls sudden release of CO2 at glacial terminations on the Chatham Rise?_ (19-UOA-339) with additional support by the _University of Auckland Doctoral Scholarship_.
Ship time was supported by the Ministry of Business, Innovation and Employment ([MBIE](https://www.mbie.govt.nz/)) and additional funding from [GNS Science](https://www.gns.cri.nz/).

## License

This project is licensed under `GNU GPLv3`. Please refer to the project [license](license.md) when considering using this workflow for your own research or project.